'use strict';
const { CrudService } = require('naf-framework-mongoose-free/lib/service');
const { BusinessError, ErrorCode } = require('naf-core').Error;
const _ = require('lodash');
const assert = require('assert');
const { ObjectId } = require('mongoose').Types;

//
class GoodsJoinActService extends CrudService {
  constructor(ctx) {
    super(ctx, 'goodsjoinact');
    this.model = this.ctx.model.Shop.GoodsJoinAct;
    this.goodsModel = this.ctx.model.Shop.Goods;
    this.goodsSpecModel = this.ctx.model.Shop.GoodsSpec;
    this.platformActModel = this.ctx.model.System.PlatformAct;
  }
  // 需要接口: 检测商品/规格 是否已经处于同种优惠方式的活动中,有就不能添加;
  // 可以存在多个不同种优惠方式的活动中: 满减,满折,赠品=> √; 满折,满折 => ×
  // 需要检查活动是否处于活动中,如果在活动中的,则不能添加
  async toCheckCanCreate(data) {
    const { spec, platform_act } = data;
    // 获取当前活动类型
    const platformAct = await this.platformActModel.findById(platform_act);
    if (!platformAct) throw new BusinessError(ErrorCode.DATA_NOT_EXIST, '未找到活动信息');
    const { type } = platformAct;
    // 找到该商品出现过的活动类型是否冲突
    const pipeline = [];
    // 找到该 规格 参与同活动类型的商品,再确定这些商品的活动是否正在进行,最后做个数量查询,如果有,就说明有冲突
    pipeline.push({ $match: { 'spec._id': spec, platform_act_type: type } });
    pipeline.push({ $addFields: { act_id: { $toObjectId: '$platform_act' } } });
    pipeline.push({
      $lookup: {
        from: 'platformAct',
        localField: 'act_id',
        foreignField: '_id',
        pipeline: [{ $match: { is_use: '0' } }],
        as: 'act',
      },
    });
    pipeline.push({ $unwind: '$act' });
    pipeline.push({ $count: 'total' });
    const list = await this.model.aggregate(pipeline);
    if (list.length > 0) throw new BusinessError(ErrorCode.SERVICE_FAULT, '该规格,已参加 正在进行中 的同类活动,不能添加');
  }

  async create(data) {
    assert(data);
    const { goods, spec, shop, platform_act } = data;
    const query = { 'goods._id': goods, 'spec._id': spec, shop, platform_act };
    const num = await this.model.count(query);
    if (num > 0) throw new BusinessError(ErrorCode.DATA_EXISTED, '该商品已添加进当前活动中');
    await this.toCheckCanCreate(data);
    // 将商品和规格 变为 快照数据
    const goodsInfo = await this.goodsModel.findById(goods);
    if (!goodsInfo) throw new BusinessError(ErrorCode.DATA_NOT_EXIST, '未找到商品信息');
    data.goods = JSON.parse(JSON.stringify(goodsInfo));
    if (spec) {
      const specInfo = await this.goodsSpecModel.findById(spec);
      if (!specInfo) throw new BusinessError(ErrorCode.DATA_NOT_EXIST, '未找到商品规格信息');
      data.spec = JSON.parse(JSON.stringify(specInfo));
    }
    const res = await this.model.create(data);
    return res;
  }

  async query(filter, { skip = 0, limit, sort, desc, projection } = {}) {
    const { platform_act, platform_act_type, shop, shop_name, goods, goods_name, goods_status, spec, spec_name, status, goods_type = 'basic' } = filter;
    const mq = {};
    const goodsMq = { status: '1' };
    if (platform_act) mq.platform_act = platform_act;
    if (platform_act_type) mq.platform_act_type = platform_act_type;
    if (shop) mq.shop = shop;
    if (goods) mq.goods = goods;
    if (goods_name) mq['goods.name'] = new RegExp(goods_name);
    if (goods_status) goodsMq.status = new RegExp(goods_status);
    if (spec) mq.spec = spec;
    if (spec_name) mq['spec.name'] = new RegExp(spec_name);
    if (status) mq.status = status;
    const pipeline = [];
    // 将加价购的加价商品过滤掉,正常列表中不显示加价购商品. 如果需要显示加价购商品,goods_type的值为plus即可
    pipeline.push({
      $match: {
        $or: [{ platform_act_type: '4', 'config.goods_type': goods_type }, { platform_act_type: { $ne: '4' } }],
      },
    });
    if (Object.keys(mq).length > 0) pipeline.push({ $match: mq });
    pipeline.push({ $addFields: { shop_id: { $toObjectId: '$shop' } } });
    const sp = [];
    if (shop_name) sp.push({ $match: { name: new RegExp(shop_name) } });
    sp.push({ $project: { name: 1 } });
    pipeline.push({
      $lookup: {
        from: 'actTags',
        localField: 'goods.act_tags',
        foreignField: 'value',
        pipeline: [{ $project: { label: 1 } }],
        as: 'goods.act_tags',
      },
    });
    pipeline.push({
      $lookup: {
        from: 'shop',
        localField: 'shop_id',
        foreignField: '_id',
        pipeline: sp,
        as: 'shopInfo',
      },
    });
    // 检查商品的实时状态及上架状态
    pipeline.push({ $addFields: { goods_id: { $toObjectId: '$goods._id' } } });
    pipeline.push({
      $lookup: {
        from: 'goods',
        localField: 'goods_id',
        foreignField: '_id',
        pipeline: [{ $match: goodsMq }],
        as: 'goods_prototype',
      },
    });
    pipeline.push({ $addFields: { gp_length: { $size: '$goods_prototype' } } });
    pipeline.push({ $match: { gp_length: { $gt: 0 } } });
    pipeline.push({
      $project: {
        meta: 1,
        shop: 1,
        shop_name: { $first: '$shopInfo.name' },
        goods: '$goods._id',
        goods_name: '$goods.name',
        // goods_status: "$goods.status",
        act_tags: '$goods.act_tags',
        spec: '$spec._id',
        spec_name: '$spec.name',
        file: '$goods.file',
        sell_money: { $toString: '$spec.sell_money' },
        flow_money: { $toString: '$spec.flow_money' },
        platform_act_type: 1,
        platform_act: 1,
        config: 1,
        status: 1,
      },
    });
    const qPipeline = _.cloneDeep(pipeline);
    qPipeline.push({ $sort: { 'meta.createdAt': -1 } });
    if (parseInt(skip)) qPipeline.push({ $skip: parseInt(skip) });
    if (parseInt(limit)) qPipeline.push({ $limit: parseInt(limit) });
    const res = await this.model.aggregate(qPipeline);
    const tPipeline = _.cloneDeep(pipeline);
    tPipeline.push({ $count: 'total' });
    const total = await this.model.aggregate(tPipeline);
    return { data: res, total: _.get(_.head(total), 'total', 0) };
  }
}

module.exports = GoodsJoinActService;
